/*
 * Copyright Bigeon Emmanuel (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * provide a generic framework for console applications.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/**
 * gclc:net.bigeon.gclc.test.utils.WritingRunnable.java
 * Created on: Nov 29, 2016
 */
package net.bigeon.gclc.utils;

/*-
 * #%L
 * Generic Command Ligne console
 * %%
 * Copyright (C) 2014 - 2018 bigeon.fr
 * %%
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * #L%
 */
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.util.ArrayDeque;
import java.util.Deque;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

/** A runnable to read the piped output.
 *
 * @author Emmanuel Bigeon */
public final class ReadingRunnable implements Runnable {

    /** The runnable to wait for arrival of a message in the queue.
     *
     * @author Emmanuel Bigeon */
    private final class ToWaitRunnable implements Runnable {
        /** The Object. */
        private final Object obj;
        /** The locking object. */
        private final Object start;
        /** The message. */
        private final String message;
        /** The started status. */
        private boolean      started = false;

        /** Create the waiting runnable.
         *
         * @param obj the object to lock on
         * @param start the object to notify when ready to wait
         * @param message the message to wait for */
        public ToWaitRunnable(final Object obj, final Object start,
                              final String message) {
            this.obj = obj;
            this.start = start;
            this.message = message;
        }

        /** Test if the waiting runnable is started.
         *
         * @return the started */
        public boolean isStarted() {
            synchronized (start) {
                return started;
            }
        }

        /* (non-Javadoc)
         * @see java.lang.Runnable#run() */
        @Override
        public void run() {
            synchronized (obj) {
                synchronized (start) {
                    started = true;
                    start.notifyAll();
                }
                while (isRunning()) {
                    try {
                        obj.wait();
                        if (delivering.equals(message)) {
                            return;
                        }
                    } catch (final InterruptedException e) {
                        LOGGER.log(Level.SEVERE, THREAD_INTERRUPTION_EXCEPTION, e);
                        Thread.currentThread().interrupt();
                    }
                }
            }
        }
    }

    /** The thread intteruption logging message. */
    private static final String THREAD_INTERRUPTION_EXCEPTION = "Thread interruption exception."; //$NON-NLS-1$
    /** The closed pipe message. */
    private static final String CLOSED_PIPE                   = "Closed pipe";                    //$NON-NLS-1$
    /** Wait timeout. */
    private static final long   TIMEOUT                       = 1000;
    /** Class logger. */
    private static final Logger LOGGER                        = Logger
            .getLogger(ReadingRunnable.class.getName());

    /** Read messages. */
    private final Deque<String> messages = new ArrayDeque<>();

    /** the reader. */
    private final BufferedReader      reader;
    /** the state of this runnable. */
    private boolean                   running            = true;
    /** Synchro object. */
    private final Object              lock               = new Object();
    /** The waiting status for a message. */
    private boolean                   waiting;
    /** The blocker for a given message. */
    private final Map<String, Object> messageBlocker     = new ConcurrentHashMap<>();
    /** The lock. */
    private final Object              messageBlockerLock = new Object();
    /** The message being delivered. */
    private String                    delivering;

    /** Create a reading runnable.
     *
     * @param reader the input to read from */
    public ReadingRunnable(final BufferedReader reader) {
        super();
        this.reader = reader;
    }

    /** do wait for a message.
     * <p>
     * This method should only be called inside a loop.
     *
     * @param timeout the timeout of the wait.
     * @throws IOException if the runnable was stopped and no essage was found. */
    private void doWaitMessage(final long timeout) throws IOException {
        try {
            synchronized (lock) {
                lock.wait(timeout);
            }
        } catch (final InterruptedException e) {
            LOGGER.log(Level.SEVERE, THREAD_INTERRUPTION_EXCEPTION, e);
            Thread.currentThread().interrupt();
        }
        if (messages.isEmpty() && !running) {
            throw new IOException(CLOSED_PIPE);
        }
    }

    /** Get the next message.
     *
     * @return the next read message
     * @throws IOException if the pipe is closed */
    public String getMessage() throws IOException {
        synchronized (lock) {
            if (!running) {
                throw new IOException(CLOSED_PIPE);
            }
            waiting = true;
            waitMessage(TIMEOUT);
            LOGGER.log(Level.FINEST, "Polled: {0}", messages.peek()); //$NON-NLS-1$
            waiting = false;
            notifyMessage(messages.peek());
            return messages.poll();
        }
    }

    /** Get the next message, but wait only a given time for it.
     *
     * @param timeout the read time out
     * @return The next message that was in the input
     * @throws IOException if the input was closed */
    public String getNextMessage(final long timeout) throws IOException {
        synchronized (lock) {
            if (!running) {
                throw new IOException(CLOSED_PIPE);
            }
            waiting = true;
            doWaitMessage(timeout);
            waiting = false;
            if (messages.isEmpty()) {
                return null;
            }
            return messages.poll();
        }
    }

    /** Get a waiting thread for a specific message delivery.
     *
     * @param message the message
     * @return the thread to join to wait for message delivery */
    public Thread getWaitForDelivery(final String message) {
        synchronized (messageBlockerLock) {
            if (!messageBlocker.containsKey(message)) {
                messageBlocker.put(message, new Object());
            }
            final Object obj = messageBlocker.get(message);
            final Object start = new Object();
            final ToWaitRunnable waitRunn = new ToWaitRunnable(obj, start, message);
            final Thread th = new Thread(waitRunn);

            // Wait for the thread to actually start before unlocking the message queue.
            synchronized (start) {
                th.start();
                while (!waitRunn.isStarted()) {
                    try {
                        start.wait(TIMEOUT);
                    } catch (final InterruptedException e) {
                        LOGGER.log(Level.SEVERE, THREAD_INTERRUPTION_EXCEPTION, e);
                        Thread.currentThread().interrupt();
                    }
                }
            }
            return th;
        }
    }

    /** Test if some data is available.
     *
     * @return if a message is waiting
     * @throws IOException if the pipe is closed */
    public boolean hasMessage() throws IOException {
        synchronized (lock) {
            if (!running) {
                throw new IOException(CLOSED_PIPE);
            }
            return !messages.isEmpty();
        }
    }

    /** Interrupts the wait on the next message by providing an empty message. */
    public void interrupt() {
        synchronized (lock) {
            if (waiting) {
                messages.offer(""); //$NON-NLS-1$
                lock.notifyAll();
            }
        }
    }

    /** Test if this element is still running.
     *
     * @return the running */
    public boolean isRunning() {
        synchronized (lock) {
            return running;
        }
    }

    /** Notify the arrival of a given message.
     *
     * @param message the message */
    private void notifyMessage(final String message) {
        synchronized (messageBlockerLock) {
            delivering = message;
            final Object mLock = messageBlocker.get(message);
            if (mLock!=null) {
                synchronized (mLock) {
                    mLock.notifyAll();
                }
                messageBlocker.remove(message);
            }
        }
    }

    /* (non-Javadoc)
     * @see java.lang.Runnable#run() */
    @Override
    public void run() {
        boolean cont = true;
        while (cont) {
            synchronized (lock) {
                cont = running;
                if (!running) {
                    break;
                }
            }
            try {
                String line = reader.readLine();
                if (line == null) {
                    // Buffer end
                    setRunning(false);
                    return;
                }
                LOGGER.log(Level.FINER, "Read: {0}", line); //$NON-NLS-1$
                line = stripNull(line);
                synchronized (lock) {
                    messages.add(line);
                    lock.notifyAll();
                }
            } catch (final InterruptedIOException e) {
                synchronized (lock) {
                    if (running) {
                        LOGGER.info("Reading interrupted"); //$NON-NLS-1$
                    }
                }
                LOGGER.log(Level.FINER, "Read interruption was caused by an exception", //$NON-NLS-1$
                        e);
            } catch (final IOException e) {
                LOGGER.log(Level.FINE, "The stream reading threw an exception", //$NON-NLS-1$
                        e);
                synchronized (lock) {
                    if (running) {
                        LOGGER.severe("Unable to read from stream"); //$NON-NLS-1$
                        setRunning(false);
                    }
                }
                return;
            }
        }
    }

    /** Set the running status for this reading runnable.
     *
     * @param running the running to set */
    public void setRunning(final boolean running) {
        synchronized (lock) {
            this.running = running;
        }
    }

    /** Wait for the next message to be integratted.
     *
     * @param timeout the timeout to wait
     * @throws IOException if the next message was not delivered and the runnable
     *         stopped. */
    private void waitMessage(final long timeout) throws IOException {
        while (messages.isEmpty()) {
            doWaitMessage(timeout);
        }
    }

    /** Strip the string from head NULL characters.
     *
     * @param line the line to strip the null character from
     * @return the resulting string */
    private static String stripNull(final String line) {
        String res = line;
        while (res.length() > 0 && res.charAt(0) == 0) {
            LOGGER.severe(
                    "NULL character heading the result of the read. This is a stream problem..."); //$NON-NLS-1$
            res = res.substring(1);
        }
        return res;
    }
}
